
package stamp.peripheral.sensor.GPS.GPSMath;

import stamp.core.*;
import com.micromegacorp.math.hardware.*;
import stamp.peripheral.sensor.GPS.Include.*;           // Contains the linked list class and other stuff

public class GPSMath
{
  Float32 m_fLat1;
  Float32 m_fLong1;
  Float32 m_fLat2;
  Float32 m_fLong2;
  Float32 m_fTemp1;
  Float32 m_fTemp2;
  Float32 m_fTemp3;
  //StringBuffer m_minutes;
  //StringBuffer m_degrees;

  public Float32 m_fTemp; // Might also be used from the outside

  // Fields can be accessed from the outside
  public Float32 m_fDistance;
  public Float32 m_fCourse;

  public GPSMath(GlobalVars globalVars)
  {
    // Create the float variables to use.
    m_fLat1 = globalVars.m_fTemp1;
    m_fLong1 = globalVars.m_fTemp2;
    m_fLat2 = globalVars.m_fTemp3;
    m_fLong2 = globalVars.m_fTemp4;
    m_fTemp = globalVars.m_fTemp5;
    m_fTemp1 = globalVars.m_fTemp6;
    m_fTemp2 = globalVars.m_fTemp7;
    m_fTemp3 = globalVars.m_fTemp8;

    m_fDistance = globalVars.m_fTemp9;
    m_fCourse = globalVars.m_fTemp10;
  }

  private boolean InitializeVariables( GPSCoordinate coordinate1, GPSCoordinate coordinate2 )
  {
    // ************** Converting to radians LAT 1 *****************************
    // We might hade something like this: 13:00,2603
    m_fLat1.set( coordinate1.m_nLatMinute ); // "Latitude 1"
    m_fTemp.set( coordinate1.m_nLatMinuteDec );
    m_fTemp.divide( 10000 );
    m_fLat1.add( m_fTemp );
    m_fLat1.divide( 60 );
    m_fLat1.add( coordinate1.m_nLatDeg );
    m_fLat1.set( Float32Math.toRadians( m_fLat1 ) );
    if (!coordinate1.m_bNorth)
      m_fLat1.negate();


    // ************** Converting to radians LONG 1 *****************************
    // Converting to radians LONG 1
    m_fLong1.set( coordinate1.m_nLongMinute ); // "Latitude 1"
    m_fTemp.set( coordinate1.m_nLongMinuteDec );
    m_fTemp.divide( 10000 );
    m_fLong1.add( m_fTemp );
    m_fLong1.divide( 60 );
    m_fLong1.add( coordinate1.m_nLongDeg );
    m_fLong1.set( Float32Math.toRadians( m_fLong1 ) );
    if (coordinate1.m_bEast)
      m_fLong1.negate();


    // ************** Converting to radians LAT 2 *****************************
    m_fLat2.set( coordinate2.m_nLatMinute ); // "Latitude 1"
    m_fTemp.set( coordinate2.m_nLatMinuteDec );
    m_fTemp.divide( 10000 );
    m_fLat2.add( m_fTemp );
    m_fLat2.divide( 60 );
    m_fLat2.add( coordinate2.m_nLatDeg );
    m_fLat2.set( Float32Math.toRadians( m_fLat2 ) );
    if (!coordinate2.m_bNorth)
      m_fLat2.negate();


    // ************** Converting to radians LONG 1 *****************************
    // Converting to radians LONG 2
    m_fLong2.set( coordinate2.m_nLongMinute ); // "Latitude 1"
    m_fTemp.set( coordinate2.m_nLongMinuteDec );
    m_fTemp.divide( 10000 );
    m_fLong2.add( m_fTemp );
    m_fLong2.divide( 60 );
    m_fLong2.add( coordinate2.m_nLongDeg );
    m_fLong2.set( Float32Math.toRadians( m_fLong2 ) );
    if (coordinate2.m_bEast)
      m_fLong2.negate();


    return ( true );
  }
  // Return crs in range 0<crs<=2*pi
  // Function changes "m_fTemp2", "m_fTemp1" and uses "crs"
  public Float32 Modcrs( Float32 crs )
  {
    // twopi = 2 * Pi
    // Modcrs = twopi - md(twopi - crs, (twopi))
    m_fTemp2.set( 2 );
    m_fTemp2.multiply( Float32Constant.pi() );

    m_fTemp1.set( m_fTemp2 );
    m_fTemp1.subtract( crs );
    m_fTemp1.set( Float32Math.mod( m_fTemp1, m_fTemp2 ) ); // now, this contains "md(twopi - crs, (twopi))"

    m_fTemp2.subtract( m_fTemp1 );
    return ( m_fTemp2 );
  }

  private Float32 CalculateCourseInternal( boolean bRecalculateDistance, int nConversionModeAngle, boolean bCheckValue )
  {
    // Sometimes when this function is called the distance is already calculated so to conserve execution time use the calculated value
    // instead of recalculating it again
    if ( bRecalculateDistance )
      m_fTemp2.set( CalculateDistanceInternal( GPSConvert.GPS_CONVERSION_METERS ) );
    else
      m_fTemp2.set( m_fDistance );

      if ( bCheckValue ) // Only if we are about to check for small value
      {
        m_fTemp3.set( "0.5" );
        if ( !( m_fTemp2.compare( m_fTemp3 ) > 0 ) )
        {
          m_fTemp.set( 0 ); // The same point
          return ( m_fTemp );
        }
    }
    // if (Abs(Cos(lat1)) < 1E-16)
    // starting point at pole
    m_fTemp.set( Float32Math.cos( m_fLat1 ) );
    m_fTemp.set( Float32Math.abs( m_fTemp ) );
    if ( !( m_fTemp2.compare( m_fTemp3 ) > 0 ) )
    {
      // gcbearing = (3 - sIn(lat1)) * Pi / 2 'starting point at pole
      m_fLat1.set( Float32Math.sin( m_fLat1 ) );
      m_fTemp2.set( 3 );
      m_fTemp2.subtract( m_fLat1 );
      m_fTemp2.multiply( Float32Constant.pi() );
      m_fTemp2.divide( 2 );

      return ( m_fTemp2 );
    }

    // gcbearing = Modcrs(Atn2(sin(lon1 - lon2) * Cos(lat2), Cos(lat1) * sin(lat2) - sin(lat1) * Cos(lat2) * Cos(lon1 - lon2)))
    m_fTemp.set( m_fLong1 );
    m_fTemp.subtract( m_fLong2 );
    m_fTemp.set( Float32Math.sin( m_fTemp ) );
    m_fTemp.multiply( Float32Math.cos( m_fLat2 ) );    // Contains "(sin(lon1 - lon2) * Cos(lat2)"

    // Calculate "sin(lat1) * Cos(lat2) * Cos(lon1 - lon2)" and store the result in "m_fTemp1"
    m_fTemp1.set( m_fLong1 );
    m_fTemp1.subtract( m_fLong2 );
    m_fTemp1.set( Float32Math.cos( m_fTemp1 ) );
    m_fTemp1.multiply( Float32Math.cos( m_fLat2 ) );
    m_fTemp1.multiply( Float32Math.sin( m_fLat1 ) ); // Contains "sin(lat1) * Cos(lat2) * Cos(lon1 - lon2)"

    m_fTemp3.set( Float32Math.sin( m_fLat2 ) );
    m_fTemp3.multiply( Float32Math.cos( m_fLat1 ) );
    m_fTemp3.subtract( m_fTemp1 );    // Contains "Cos(lat1) * sin(lat2) - sin(lat1) * Cos(lat2) * Cos(lon1 - lon2)"

    // Reusing the variable "m_fLat1" since it's not longer used
    m_fLat1.set( Float32Math.atan2( m_fTemp, m_fTemp3 ) );

    m_fTemp.set( Modcrs( m_fLat1 ) );

    // Only if we're converting to degrees
    if ( nConversionModeAngle == GPSConvert.GPS_CONVERSION_DEGREES )
      m_fTemp.set( Float32Math.toDegrees( m_fTemp ) );

    return ( m_fTemp );

  }
  /*
  // This function is converting the result "fValue" to the conversion specified by the "nConversionMode"
  // The result is the formatted string containing the correct result
  public static boolean ConvertGPSResult( Float32 fValue, int nConversionMode, StringBuffer buffer, Float32 fTemp )
  {
    int nResultFormat = 0;
    buffer.clear();

    // If no conversion, exit
    if ( nConversionMode == GPSConvert.GPS_CONVERSION_NONE )
    {
      buffer.append( fValue.floatFormat( nResultFormat ) );
      return ( true );
    }

    // Assume conversion to nautical miles
    fValue.multiply( 10800 ); // The same as 180x60
    fValue.divide( Float32Constant.pi() );
    if ( (nConversionMode & GPSConvert.GPS_CONVERSION_NAUTICALMILES) > 0 ) // Nothing to do anymore
    {
      if ( ( nConversionMode & GPSConvert.GPS_CONVERSION_SCALE_AUTO ) > 0 )
      {
        nResultFormat = 91;

        // Please see the GPSConvert class for details
        if ( fValue.compare( 1 ) < 0 )
        {
          nResultFormat = 92;

          // Check if the value is less than 0.1 nm, in that case we return the answer in meters
          fTemp.set( "0.1" ); // one cable length
          boolean bConvertToMeters = false;

          if ( fValue.isNegative() )
          {
            fTemp.negate();
            bConvertToMeters = ( fValue.compare( fTemp ) >= 0 );
          }
          else
            bConvertToMeters = ( fValue.compare( fTemp ) < 0 );

          if ( bConvertToMeters )
          {
            // Less than 0.1 cable lengths
            fValue.multiply( 1852 );
            //fValue.subtract( 44 );
            buffer.append( fValue.floatFormat( 50 ) );
            buffer.append( "m");
            return ( true );
          }
        }
      }

      buffer.append( fValue.floatFormat( nResultFormat ) );
      buffer.append( "nm" );
      return ( true );
    }

    // Comming to this point we wither have meters or kilometers, so we calculate
    // the answer in meters and if the result should be in kilometers, then we divide the
    // answer by 1000

    // Assume meters as default
    fValue.multiply( 1852 );
    //fValue.subtract( 44 );      // Why this error?? Always of 44 meters!
    nResultFormat = 90; // We round the result and return the rounded value without any decimals

    // Answer in kilometers
    if ( (nConversionMode & GPSConvert.GPS_CONVERSION_KILOMETERS ) > 0 )
    {
      nResultFormat = 91;

      fTemp.set( 1000 ); // One kilometer in meters
      if ( fValue.isNegative() )
        fTemp.negate();

      // Now, check if less than 1000 km - if so, the result is in meters
      if ( fValue.compare( fTemp ) > 0 )
      {
        fValue.divide( 1000 ); // The result is in km
        buffer.append( fValue.floatFormat( nResultFormat ) );
        buffer.append( "km");
        return ( true );
      }
    }

    // The result is in m
    buffer.append( fValue.floatFormat( nResultFormat ) );
    buffer.append( "m");

    // Return the result with the formatting we decided previously
    return ( true );
  }
  */
  // This function calculates both the course and the distance and returns "true" if everything is OK or "false" if either
  // the coordinates could not be parsed or the calculations failed
  public boolean CalculateCourseAndDistance( GPSCoordinate coordinate1, GPSCoordinate coordinate2,
                                             Float32 fCourse, Float32 fDistance,
                                             int nConversionModeCourse, int nConversionModeDistance )
  {
    // Initialize the variables used in this class and convert the input to radians
    if ( !InitializeVariables( coordinate1, coordinate2 ) )
      return ( false );

    // This is the distance
    m_fDistance.set( CalculateDistanceInternal( nConversionModeDistance ) );
    fDistance.set( m_fDistance );

    m_fTemp.set( fDistance );

    // Assume we have meters
    boolean bComputeCourse = ( m_fTemp.compare( 4 ) >= 0 ); // We can compute the course between 2 points if they are more than 3 meters apart

    // Now, find out if we can calculate the course - if we are too near, we can't
    if ( ( nConversionModeDistance & GPSConvert.GPS_CONVERSION_NONE ) > 0 )
    {

      m_fTemp.multiply( 10800 ); // The same as 180x60
      m_fTemp.multiply( 1852 ); // Convert the distance to meters
      bComputeCourse = ( m_fTemp.compare( 4 ) >= 0 ); // We can compute the course between 2 points if they are more than 3 meters apart
    }
    else if ( ( nConversionModeDistance & GPSConvert.GPS_CONVERSION_NAUTICALMILES ) > 0 )
    {
      m_fTemp.multiply( 1852 ); // Convert the distance to meters
      bComputeCourse = ( m_fTemp.compare( 4 ) >= 0 ); // We can compute the course between 2 points if they are more than 3 meters apart
    }
    else if ( ( nConversionModeDistance & GPSConvert.GPS_CONVERSION_KILOMETERS ) > 0 )
    {
      m_fTemp.multiply( 1000 ); // Convert the distance to meters
      bComputeCourse = ( m_fTemp.compare( 4 ) >= 0 ); // We can compute the course between 2 points if they are more than 3 meters apart
    }

    // Only if we can compute the course
    if ( bComputeCourse )
      fCourse.set( CalculateCourseInternal( false, nConversionModeCourse, false ) );

    return ( true );
  }

  /*
  // Function calculates the course between 2 locations on earth, i.e. the course from (Lat1, Long1) to (Lat2, Long2)
  public boolean CalculateCourse( GPSCoordinate coordinate1, GPSCoordinate coordinate2, Float32 fCourse )
  {
    // Initialize the variables used in this class and convert the input to radians
    if ( !InitializeVariables( coordinate1, coordinate2 ) )
      return ( false );

    // This is the calculated course
    fCourse.set( CalculateCourseInternal( true, GPSConvert.GPS_CONVERSION_DEGREES, true ) );

    return ( true );
  }
  */


  // Calculates the distance between 2 coordinates on earth
  // Input format (example):
  // Point 1: 55 (degrees), 13.55 (minutes), 12 (degrees), 10.2 (minutes)
  // Point 2: 56 (degrees), 1 (minutes), 10 (degrees), 10.2 (minutes)
  public boolean CalculateDistance( GPSCoordinate coordinate1, GPSCoordinate coordinate2, Float32 fDistance, int nConversionMode )
  {
    // Initialize the variables used in this class and convert the input to radians
    if ( !InitializeVariables( coordinate1, coordinate2 ) )
      return ( false );

    // This is the calculated distance
    fDistance.set( CalculateDistanceInternal( nConversionMode) );

    return ( true );
  }
  /*
  Function Acs(X)
    ' Equivalent of spreadsheet Acos(x)
    If (X >= 0) Then
       Acs = 2 * Atn(Sqr((1 - X) / (1 + X)))
    Else
      Acs = Pi - 2 * Atn(Sqr((1 + X) / (1 - X)))
    End If
  End Function
*/

  /*
  // This function is not using m_fTemp, but m_fTemp1 and m_fTemp2
  private Float32 Acs( Float32 fX )
  {
    m_fTemp1.set( 1 );

    // If it's positive
    if ( !fX.isNegative() )
    {
      m_fTemp1.subtract( fX );

      fX.add( 1 );
      m_fTemp1.divide( fX );

      fX.set( Float32Math.sqrt( m_fTemp1 ) );
      fX.set( Float32Math.atan( fX ) );
      fX.multiply( 2 );
    }
    else
    {
      m_fTemp1.add( fX );

      m_fTemp2.set( 1 );
      m_fTemp2.subtract( fX );

      m_fTemp1.divide( m_fTemp2 );
      fX.set( Float32Math.sqrt( m_fTemp1 ) );
      fX.set( Float32Math.atan( fX ) );
      fX.multiply( 2 );

      m_fTemp1.set( Float32Constant.pi() );
      fX.set( m_fTemp1.subtract( fX ) );
    }

    return ( fX );
  }
  */
  // This function is not using m_fTemp
   private Float32 Asn( Float32 fX )
   {
    // Asn = 2 * Atn(X / (1 + Sqr(1 - X * X)))
    m_fTemp1.set( 1 );
    m_fTemp2.set( Float32Math.pow( fX, 2 ) );

    m_fTemp1.subtract( m_fTemp2 );
    m_fTemp1.set( Float32Math.sqrt( m_fTemp1 ) );
    m_fTemp1.add( 1 );

    fX.divide( m_fTemp1 );
    fX.set( Float32Math.atan( fX ) );
    fX.multiply( 2 );

    return ( fX );
   }
   private Float32 CalculateDistanceInternal( int nConversionMode )
   {
    // (sin((lat1 - lat2) / 2)) ^ 2
    m_fTemp1.set( m_fLat1 );
    m_fTemp1.subtract( m_fLat2 );
    m_fTemp1.divide( 2 );
    m_fTemp1.set( Float32Math.sin( m_fTemp1 ) );
    m_fTemp1.set( Float32Math.pow( m_fTemp1, 2 ) );

    // cos(lat1) * cos(lat2)
    m_fTemp2.set( Float32Math.cos( m_fLat1 ) );
    m_fTemp3.set( Float32Math.cos( m_fLat2 ) );
    m_fTemp2.multiply( m_fTemp3 );

    // (sin((lon1 - lon2) / 2)) ^ 2
    m_fTemp3.set( m_fLong1 );
    m_fTemp3.subtract( m_fLong2 );
    m_fTemp3.divide( 2 );
    m_fTemp3.set( Float32Math.sin( m_fTemp3 ) );
    m_fTemp3.set( Float32Math.pow( m_fTemp3, 2 ) );

    // cos(lat1) * cos(lat2) * (sin((lon1 - lon2) / 2)) ^ 2
    m_fTemp2.multiply( m_fTemp3 );

    // (sin((lat1 - lat2) / 2)) ^ 2 +  cos(lat1) * cos(lat2) * (sin((lon1 - lon2) / 2)) ^ 2
    m_fTemp1.add( m_fTemp2 );

    m_fTemp.set( Float32Math.sqrt( m_fTemp1 ) );

    m_fTemp.set( Asn( m_fTemp ) );
    //m_fTemp.set( Float32Math.asin( m_fTemp ) );
    m_fTemp.multiply( 2 );

    if ( m_fTemp.isNegative() )
    {
      m_fTemp.set(0); // This is not good...the calculation failed
      return ( m_fTemp );
    }

    // No conversion, return the radian value
    if ( ( nConversionMode & GPSConvert.GPS_CONVERSION_NONE ) > 0 )
      return ( m_fTemp );

    // There is some kind of conversion, so assume [nm]
    m_fTemp.multiply( 180 );
    m_fTemp.multiply( 60 );
    m_fTemp.divide( Float32Constant.pi() );

    // Answer in meters
    if ( ( nConversionMode & GPSConvert.GPS_CONVERSION_METERS ) > 0  )
    {
      m_fTemp.multiply( 1852 );

      // Why this error?? Always of 44 meters!
      m_fTemp.subtract( 44 );
    }

    // Answer in kilometers
    if ( ( nConversionMode & GPSConvert.GPS_CONVERSION_KILOMETERS ) > 0 )
    {
      m_fTemp2.set("1.852");
      m_fTemp.multiply( m_fTemp2 );
    }

    return ( m_fTemp );
   }
}