
package stamp.peripheral.sensor.GPS.Include;

import stamp.core.*;
import stamp.peripheral.sensor.GPS.RouteFollower.*;
import stamp.peripheral.memory.eeprom.*;



public class Settings
{
  // For the route follower
  public int m_nRoute;
  public int m_nWayPoint;
  public int m_nFollowMode;

  // For the time offset
  public int m_nTimeOffset;
  public int m_nBackLightMode;

  public void ReadSettings(MC24LC256 eeprom)
  {

    //*****************************************************************************************//
    //                            Route follower                                               //
    //*****************************************************************************************//
    boolean b = true;
    int nAddress = MemoryRouteHandler.SETTINGS_START;
    try
    {
      m_nRoute = eeprom.readRandom( MemoryRouteHandler.m_dev, nAddress );       // Read the route

      nAddress++;
      m_nWayPoint = eeprom.readRandom( MemoryRouteHandler.m_dev, nAddress );    // Read the waypoint

      nAddress++;
      m_nFollowMode = eeprom.readRandom( MemoryRouteHandler.m_dev, nAddress );  // Read the follow mode
    }
    catch(Exception ex)
    {
      b = false;
    }
    if (!b)
    {
      m_nRoute = 0;
      m_nWayPoint = 0;
      m_nFollowMode = RouteFollower.FOLLOW_NONE;
    }

    nAddress = MemoryRouteHandler.SETTINGS_START + 3;
    //*****************************************************************************************//
    //                            Time offset                                                  //
    //*****************************************************************************************//
    try
    {
      // Byte 1
      nAddress++;
      m_nTimeOffset = eeprom.readRandom( MemoryRouteHandler.m_dev, nAddress ) << 8;

      // Byte 2
      nAddress++;
      m_nTimeOffset |= eeprom.readRandom( MemoryRouteHandler.m_dev, nAddress );

    }
    catch(Exception ex)
    {

    }

    nAddress = MemoryRouteHandler.SETTINGS_START + 6;
    //*****************************************************************************************//
    //                            Backlight handler                                            //
    //*****************************************************************************************//
    try
    {
      nAddress++;
      m_nBackLightMode = eeprom.readRandom( MemoryRouteHandler.m_dev, nAddress );
    }
    catch(Exception ex)
    {

    }

    nAddress = MemoryRouteHandler.SETTINGS_START + 8;
  }


  public void SaveSettings(MC24LC256 eeprom)
  {

    //*****************************************************************************************//
    //                            Route follower                                               //
    //*****************************************************************************************//
    int nAddress = MemoryRouteHandler.SETTINGS_START;
    eeprom.writeVerify( MemoryRouteHandler.m_dev, nAddress, (byte)(m_nRoute&0x00FF) ); // The current route

    nAddress++;
    eeprom.writeVerify( MemoryRouteHandler.m_dev, nAddress, (byte)(m_nWayPoint&0x00FF) ); // The current waypoint

    nAddress++;
    eeprom.writeVerify( MemoryRouteHandler.m_dev, nAddress, (byte)(m_nFollowMode&0x00FF) ); // The follow mode

    m_nTimeOffset = -12;
    //*****************************************************************************************//
    //                            Time offset                                                  //
    //*****************************************************************************************//
    // Byte 1
    nAddress++;
    byte _byte = (byte)(( m_nTimeOffset >> 8 ) & 0x00FF );
    eeprom.writeVerify( MemoryRouteHandler.m_dev, nAddress, _byte);

    // Byte 2
    nAddress++;
    _byte = (byte)(m_nTimeOffset & 0x00FF);
    eeprom.writeVerify( MemoryRouteHandler.m_dev, nAddress, _byte);
    //*****************************************************************************************//
    //                            Backlight handler                                            //
    //*****************************************************************************************//
    nAddress++;
    _byte = (byte)(m_nBackLightMode & 0x00FF);
    eeprom.writeVerify( MemoryRouteHandler.m_dev, nAddress, _byte);
  }
}